# photcheck -- monitor photometric stability of a data set.
#
# The user first identifies several suitable objects in a mosaiced image using
# imexam.  These objects should be located on a portion of the mosaic with 
# nearly maximum exposure time, so that they will appear in most or all of 
# the individual input frames.  Photcheck then takes that input position list, 
# transforms it back to the coordinate systems of the individual sky subtracted
# images which were summed to make the mosaic, and uses digiphot.apphot.phot to 
# compute accurate positions and measure magnitudes for the reference objects 
# in each sky subtracted input frame.   
#
# If the user sets badcheck=yes, then each image's bad pixel mask is consulted,
# and if any bad pixels are within the photometric aperture, phot will return
# an INDEF value for the magnitude.   Finally, the output files from phot are 
# compiled using awk into a format suitable for input to "photscale", which 
# allows interactive calculation of photometric scalings for the data set.
#
# The default output uses the mosaic image as the reference photometric image.
# If another image is desired as a reference, that can be done in `photscale'.
#
# This routine requires: minv, fileroot, photcheck.awk; and that awk be 
# defined in the user package as a foreign task, i.e. task $awk = $foreign.
#
# 930803: originally created 2-3 Aug 1993  by Mark Dickinson
# 960321: modified to include interactive identification of ref. stars, etc.
# 960326: include editing of imshifts file, and check for overlong image names
# 960409: changed _badpix.imh to unique filename, & removed duplicate creation
# 960410: fixed bad pixel image bug
# 960418: allowed for flipping of coadded image
# 960420: removed dephot option in favor of editing PHOT parameter set
# 960716: allow for multiple phot apertures by retrieving only mag[1] & merr[1]
# 980511: fixed minor bug where "ext" could be called on w/o being defined
# 990225: minor change to calling of "phot"
# 990513: use imcalc now instead of imar
# 990714: intermediate file for awk to work on more platforms

procedure photcheck(image,expmap,photcoords,output)

string	image  		{prompt="Mosaic image"}
string	expmap  	{prompt="Exposure map image"}
string	photcoords	{prompt="File for reference star coordinates"}
string	output		{prompt="File for formatted output photometry"}

string	imageshifts=""	{prompt="`imshifts' file (default = image//.imshifts)"}
bool	flipped     {no, prompt="Was coadded image flipped during ircoadd?"}
bool	interact   {yes, prompt="Identify reference stars interactively?"}
bool	ephot      {yes, prompt="Edit PHOT parameter sets before beginning?"}
bool	badcheck   {yes, prompt="Check for bad pixels and return INDEF values?"}
bool	verbose	   {yes, prompt="Verbose output?"}
bool	cleanup	   {yes, prompt="Erase individual photometry files when done?"}
bool   echog        {yes,prompt="Beep when task is done?"}
struct	*imglist

begin

	string 	mimg,photc,outp	# Equal query parameters
	int 	oxdim,oydim	# Dimensions of mosaic image
	int	ix, iy		# Coordinate offsets of input images
	string	img		# Fscanned name of image
	string	ext		# extension of image (imh, hhh, etc.)
	string	starshift	# File for translated star coordinates
	string	workimage	# Copy of image for use with phot.
	real	xstar, ystar	# Star coordinates
	real	temp		# swap variable
	string	photlist	# temporary dump file
	string	phiirsdir	# pathname of PHIIRS source code directory
	real	flagval		# large flag value -- set to 1.e9
	string	mexp, bpmold, bpmnew, imagelist, s1, imgroot, morig, tempbp
	string  old_exposure, old_gain, old_ccdread, old_apertures, old_filter
	real	old_epadu, old_readnoise, old_annulus, old_dannulus, old_zmag
	real	old_scale
	int	x, y, z
	bool	iinteract, eephot, fflipped

	flagval = -1.e9

# Check to make sure digiphot.apphot is loaded.
if (!defpac("apphot") || !defpac("imgtools")) {
	beep
	print ("Please load digiphot.apphot and/or stsdas.imgtools before using this routine.")
	return
}

# Check for presence of the photcheck.awk script in the PHIIRS source 
# code directory.  If that file is not present, return error message.
phiirsdir = osfn("phiirs$")
if (!access(phiirsdir//"photcheck.awk")) {
	beep
	print ("ERROR:   Cannot access "//phiirsdir//"$photcheck.awk")
	return
}

# Check for presence tasks needed to run script
if (!deftask("minv") || !deftask("fileroot")) {
	beep
	print ("ERROR:  tasks `minv' and/or `fileroot' are not loaded.")
	print ("Please load them and restart task.")
	return
}

# Get query parameters.
fileroot(image,validim+)
	mimg = fileroot.root
	morig = mimg
mexp  = expmap
	if(mexp=="")mexp=mimg//"em"
photc = photcoords
	if(photc=="")photc=mimg//".photcoords"
outp  = output
	if(outp=="")outp=mimg//".photcheck"
	if (access(outp)) {
		beep
   print ("ERROR: File "//outp//" already exists.  Choose another output name.")
		return
	}

imagelist = imageshifts
	if(imagelist=="")imagelist=mimg//".imshifts"
fflipped = flipped
	if(fflipped){
		if(access(mimg//"_F.imh"))imdel(mimg//"_F", ver-)
		if(access(mimg//"_F.hhh"))imdel(mimg//"_F", ver-)
		imcopy(mimg//"[-*,*]", mimg//"_F", ver-)
		mimg=mimg//"_F"
		if(strlen(mimg)>=22){
			print("WARNING!  Image name "//mimg//" is too long!")
			print("Image names can't be longer than 21 characters.")
			print("Please rename image(s) and restart the task.")
			return
		}
	}
ext=""

# Check to make sure no image names are more than 21 characters long
imglist=imagelist
while (fscan(imglist,img,ix,iy) != EOF) {
	if(strlen(img)>=22){
		beep
		print("WARNING!  Image name "//img//" is too long!  Image names as listed in")
		print(imagelist//" can be no longer than 21 characters.")
		print("Please rename image(s) and restart the task.")
		return
	}
}
imglist=""
iinteract = interact
eephot = ephot
	if(!eephot)print("PHOT task parameter sets will not be edited.")

# Display mosaic image for reference star identification, if desired
if(iinteract){
print("===> The mosaic image and exposure map are being displayed.  Identify") 
print("===> several objects suitable for photometry in the mosaic image,")
print("===> preferably ones located on parts of the mosaic with nearly")
print("===> maximum exposure time, and get their coords (type 'a').")
print("")
print("===> When done, the coordinate file can be edited in case mistakes were made.")
print("===> When using preexisting coordinate file, this temporary file will be blank.")
print("")
print("===> Finally, the imshifts file will be displayed for editing.  Be sure the")
print("===> listed images were the actual ones coadded to produce the mosaic.")
	if(fflipped){display(mexp//"[-*,*]",2)}
		else{display(mexp,2)}
	display(mimg,1)
	imexam(logfile="tmp"//photc, keep+)
        vi("tmp"//photc)
        vi(imagelist)
# Output reference star x and y values to desired file
   if(fflipped)print("# WARNING! Coordinates are for "//morig//"[-*,*]",>>photc)
	fields("tmp"//photc,"1-2",lines="1-99999",>>photc)
	del("tmp"//photc,ver-)
}

# Get dimensions of mosaic image.
imgets(mimg,"i_naxis1")
oxdim=int(imgets.value)
imgets(mimg,"i_naxis2")
oydim=int(imgets.value)

# If bad pixel checking is turned on, delete temporary files if present.
if (badcheck) {
# datapars.datamin is set to trigger 'INDEF' results for any object which has
# a bad pixel within the aperture.  If badcheck=no, then leave datamin = INDEF.
	datapars.datamin = 0.95 * flagval
   } else {
	datapars.datamin = INDEF
}

# If desired, set up default values for important PHOT parameters
if(eephot){
	epar("datapars")
	epar("fitskypars")
	epar("photpars")
}

# Create temporary file for phot output filenames.
photlist = mktemp ("photcheck")

# First, do photometry on stars in the mosaic image to act as reference values.
if (verbose) print ("Measuring reference magnitudes from ",morig)
if(access(morig//".mag.X")){
	#del(morig//".mag.X",ver-)
	print("WARNING! "//morig//".mag.X already exists!  Not overwritten.")
} else {
	phot (mimg,skyfile="",datapars="",centerpars="",fitskypars="",photpars="",coords=photc,output=morig//".mag.X",inter-,verify-,verbose-)
}

# Record photometry output file name in temporary file photlist.
print (morig//".mag.X", >photlist)

# Next, loop through the individual images.
bpmold=""
bpmnew=""
imglist=imagelist
while (fscan(imglist,img,ix,iy) != EOF) {
	fileroot(img,validim+)
	if (badcheck) {
		workimage = "_"//img
		imgroot="_"//fileroot.root
	} else {
		workimage = img
		imgroot=fileroot.root
	}

# If photometry files already exist, do not re-create them
	if(access(imgroot//".mag.X")){
	print("WARNING! "//imgroot//".mag.X already exists! Not overwritten.")
	} else {
# If photometry files do not exist, create them.  Used to rename bad pixel mask 
# to *.imh, thus giving it datatype=real to allow large flag values used later,
# but now I just use 1.e9 as the large flag value, which works for *.pl images.
	if (verbose) print ("Working on image ",img)
	if (badcheck) {
		imgets (img,"BPMASK")
		bpmnew=imgets.value
		fileroot(bpmnew,validim+)
		tempbp=fileroot.root
		ext=fileroot.extension
		if(ext!="")ext="."//ext
## Only create new bad pixel mask if image BPM is different than previous image
#		if(tempbp!=bpmold){
#			imdel ("_badp"//bpmold//ext,ver-,>&"dev$null")
#			imdel ("_newp"//bpmold//ext,ver-,>&"dev$null")
#			imdel ("_badp"//tempbp//ext,ver-,>&"dev$null")
#			imdel ("_newp"//tempbp//ext,ver-,>&"dev$null")
#	                imcopy (tempbp,"_badp"//tempbp//ext,ver-)
#	               minv("_badp"//tempbp//ext,"_newp"//tempbp//ext)
#      imar("_newp"//tempbp//ext,"*",flagval,"_newp"//tempbp//ext,ver-)
#		}
## Copy image and set bad pixels to a large, negative flag value.
#	        imar(img,"*","_badp"//tempbp//ext,"_"//img,ver-,noact-)
#		imar("_"//img,"-","_newp"//tempbp//ext,"_"//img,ver-)
#		bpmold=tempbp
# Set bad pixels to a large, negative flag value.
	        imcalc(img//","//tempbp, "_"//img, "if im1 .eq. 0 then "//flagval//" else im1", nullval=flagval, ver-)
	}

# Apply translation to star coordinate file.
	starshift = mktemp("photcheck")
	lintran (photc,xf=1,yf=2,x1=ix,y1=iy,x2=0.,y2=0.,xsc=1.,ysc=1.,angle=0,min_sig=4,radians=no,>>starshift)

# Do photometry on star list.   Output goes to file named imgroot//.mag.X
	phot (workimage,skyfile="",datapars="",centerpars="",fitskypars="",photpars="",coords=starshift,output=imgroot//".mag.X",inter-,verify-,verbose-)

# End of photometry file creation loop.  Clean up.
	delete (starshift,ver-)
	if (badcheck) {
		imdelete ("_"//img,ver-)
	}
}

# Record photometry output file name in temporary file photlist.
	print (imgroot//".mag.X",>>photlist)

# Next image
}
imdelete ("_badp"//bpmnew//ext,ver-,>&"dev$null")
imdelete ("_newp"//bpmnew//ext,ver-,>&"dev$null")

# Dump photometric records to output photometry file & format using awk scripts
#txdump ("@"//photlist,"image,xairmass,mag[1],merr[1]","yes",headers+,parameters-) | awk ("-f",phiirsdir//"photcheck.awk",> outp)
del ("_tmpoutp", ver-, >& "dev$null")
txdump ("@"//photlist,"image,xairmass,mag[1],merr[1]","yes",headers+,parameters-, >> "_tmpoutp")
awk ("-f", phiirsdir//"photcheck.awk", "_tmpoutp", > outp)
print ("", >> outp)

if (verbose) {
	print ("")
	print ("Output data written to file "//outp//"; now, if desired,")
	print ("replace the first line in it with the desired reference")
	print ("photometry values, and then use it as input to task photscale")
	print ("to compute and plot photometric offsets between frames.")
}

# Clean up.
if(fflipped) imdelete (mimg,ver-)
imdelete ("_badp*"//ext,ver-,>&"dev$null")
imdelete ("_newp*"//ext,ver-,>&"dev$null")
if(cleanup){
	delete ("@"//photlist,ver-)
	delete (photlist,ver-)
	delete ("_tmpoutp", ver-)
} else {
	if(!access(morig//photlist)){
	      if(!access(morig//".photlist"))rename(photlist,morig//".photlist")
	     print("Output photometry files are listed in "//morig//".photlist")
	} else {
		print("Output photometry files are listed in "//photlist)
        }
}

imglist = ""
if(echog)beep

end
