# qkflat.cl -- quick IR flattening routine using all images in a list
#
# This task creates a flat from all images in the input list and flattens
# the images in the list with it.
#
# The task should be run in the directory which contains the raw images *.###
# listed in imagelist.  This script does not change the original images.  
# Previously a bad pixel image (mask) should have been produced.  
#
# darksub=no:
#       imagelist: input
#       flatlist: imagelist or imagelist * bpmask if zeroout+
#       flatmakelist: imagelist / avg(flatlist)
#       imcom @flatmakelist flat
#       flatimagelist: imagelist / flat
# darksub=yes:
#       imagelist: input
#       flatimagelist: imagelist - dark
#       flatlist: flatimagelist or flatimagelist * bpmask if zeroout+
#       flatmakelist: flatimagelist / avg(flatlist)
#       imcom @flatmakelist flat
#       flatimagelist: flatimagelist / flat
#
# CHANGES:
# 940729: images used in creating the flat are normalized to their
# 	average sky level before the flat is made... seems like this should
# 	have been a bug in earlier flats... problem w/older imcombine?
# 940912: speeded things up 
# 940917: added compatibility with iraf version 2.10, in which ccdclip in 
#	imcombine doesn't work; and added stacking capability for making flat.
# 981230: fixed minor bug which showed up when image extension was "fits"
# 990115: added option to just make the flat and not apply it to the images
# 990217: added option to use iterstat; this slows things down if used but
#	produces a slightly better flat
# 990218: added mkflat option, fixing prob of nondeletion of flatbase to boot
# 990225: fixed darksub+ bug; moved zeroout+ from all output images to just flat
#	(+ divzero=0. during flat division); don't subtract dark if already done
# 	and do iterstat such that results propagate through to the final images
# 990306: add DARKSUB header card after dark subtraction
# 990317: added tpl option
# 990419: minor bug in flat normalization fixed ... no big effect
# 990506: added ZEROOUT keyword creation
# 990523: set scale="none" during (im)combine step
# 010315: changed "access" to "imaccess" in mkflat check

procedure qkflat(rootname,imagelist,flatstring,fextension,logfile,darksub,dark,badmask,zeroout,badpix)

string rootname   {prompt="Rootname for flatfield image"}
string imagelist  {prompt="List of images to process (do not use an `@')"}
string flatstring {"fq",prompt="String to be appended to flattened images"}
string fextension {"fq",prompt="Extension to be added to the flatfield image"}
string logfile    {prompt="Name of log file"}
bool   darksub=no {prompt="Subtract Dark frame? (yes|no)"}
string dark=""    {prompt="Name of the Dark frame to use"}
string badmask="" {prompt="Name of the Bad Pixel Mask to use"}
bool   zeroout=no {prompt="Zero out bad pixels on final images? (yes|no)"}
bool   badpix=no  {prompt="Add bad pixel mask name to image header? (yes|no)"}

bool   mkflat=no  {prompt="Make flatfield image even if it already exists?"}
bool iterstat=yes {prompt="Use iterative means when constructing flats?"}
bool   doflat=yes {prompt="Use the flatfield to create flattened images?"}
bool   tpl=no     {prompt="Throughput-correct images after flatfielding?"}
string tplname    {prompt="FULL Name of throughput correction image"}
bool   v210=no    {prompt="Older version of IRAF w/imcombine bug? (yes|no)"}
pset   imflatpars {prompt="Imcombine parameters"}
bool   echog=yes  {prompt="Beep when task is done?"}
struct *list
struct *list2
struct *datelist
struct *meanlist

begin

string s_sigma, s_combine, s_reject, s_offsets, s_masktyp, s_gain, s_rdnoise
string s_scale, s_zero, s_weight, s_statsec, s_expname, templist, postdsimage
string rrootname, ddark, llogfile, bbadmask, eextension, ffextension, predsimage
string flatlist, flat, flatbase, meanfile, currentflat, tempflat, origimage
string iimagelist, flatimagelist, fflatstr, flatmakelist, flatjoinlist, newflat
string datefile, thedate, ickinfo, name, base, check, tempstring, otherstring
string ttplname
real mean, meaner, meanest, meanbad, x, y, z
real s_maskval, s_blank, s_lthresh, s_hthresh, s_nlow, s_nhigh, s_nkeep
real s_lsigma, s_hsigma, s_pclip, s_grow
int sstartno, eendno, i, j, numpix
bool ddarksub, zzeroout, bbadpix, go, ddoflat, iter_stat, mmkflat, ttpl

rrootname = rootname
iimagelist = imagelist
list = iimagelist
fflatstr = flatstring
ffextension = fextension
llogfile = logfile
ddarksub = darksub
ddark = dark
bbadmask = badmask
zzeroout = zeroout
bbadpix = badpix

mmkflat = mkflat
iter_stat = iterstat
ddoflat = doflat
ttpl = tpl
	if (ttpl) ttplname=tplname

s_sigma = i_sigma
s_combine = i_combine
s_reject = i_reject
s_offsets = i_offsets
s_masktyp = i_masktyp
s_maskval = i_maskval
s_blank = i_blank
s_scale = i_scale
s_zero = i_zero
s_weight = i_weight
s_statsec = i_statsec
s_expname = i_expname
s_lthresh = i_lthresh
s_hthresh = i_hthresh
s_nlow = i_nlow
s_nhigh = i_nhigh
s_nkeep = i_nkeep
s_lsigma = i_lsigma
s_hsigma = i_hsigma
s_rdnoise = i_rdnoise
s_gain = i_gain
s_pclip = i_pclip
s_grow = i_grow

# Set up log file, printing warning if it already exists
# The "go" parameter won't let the task run if you're using "combine"
# instead of "imcombine" and haven't loaded the ccdred package.
go=yes
if(!defpac("ccdred") && v210) {go=no}
if(go) {
print("Let the games begin...")
if(llogfile == "") llogfile=rrootname//".log"
if(access(llogfile))
  print("Log file `"//llogfile//"' already exists... will append to it.")
print(" ",>>llogfile);print("PHIIRS.QKFLAT-------------------------",>>llogfile)

# Set up files
flat = rrootname//"."//ffextension    		 # the final flat field image
flatimagelist = iimagelist//fflatstr  	 	 # @file for flattened images
flatmakelist = "listmake."//rrootname//fflatstr  # @file for making the flat
flatjoinlist = "listjoin."//rrootname//fflatstr  # @file for making the flat
flatlist = mktemp("flatlist")         		 # @file for temp flat files
flatbase = mktemp("flat")             		 # temp flat files base name
if(access(flatimagelist)) delete(flatimagelist,ver-)
if(access(flatmakelist)) delete(flatmakelist,ver-)
while (fscan (list, name) !=EOF) {
  fileroot(name,validim+)
  check=fileroot.extension
  if(check!="")check="."//check
  base=fileroot.root
  print(base//fflatstr//check, >> flatimagelist)
  print(base//fflatstr//check, >> flatmakelist)
  print(flatbase//base//fflatstr//check, >> flatlist)
}

# Make flat, if desired
if (mmkflat || !imaccess(flat)) {
print ("===> Making flat...")
if (imaccess(flat)) imdel (flat, ver-)
print("File "//flatlist//" contains the following filenames:", >> llogfile)
type(flatlist, >> llogfile); print(" ", >> llogfile)
print("File "//flatimagelist//" contains the following filenames:", >> llogfile)
type(flatimagelist, >> llogfile); print(" ", >> llogfile)

# Add the BPM to the image header
if(bbadpix){
	hedit("@"//iimagelist,"BPM",bbadmask,add+,ver-,show-,update+)
}

# If desired, subtract off Dark Current in both flatlist and flatimagelist
if (ddarksub) {
	print("===> Subtract off Dark Current")
	print("===> Subtract off Dark Current", >> llogfile)
	join (iimagelist, flatimagelist, out=flatjoinlist)
	list = flatjoinlist
	while (fscan (list, predsimage, postdsimage) !=EOF) {
		imgets(predsimage,"darksub",>&"dev$null")
        	if(imgets.value=="0"){
			imarith (predsimage,"-",ddark,postdsimage,ver-,noact-)
			hedit(postdsimage,"DARKSUB",ddark,add+,ver-,sho+,updat+)
		  print(predsimage//" - "//ddark//" "//postdsimage, >> llogfile)
		} else {
			imcopy (predsimage, postdsimage, ver-)
			print (predsimage//" already darksubtracted; just copied to "//postdsimage, >> llogfile)
			print (predsimage//" already darksubtracted; just copied to "//postdsimage)
		}
	}
	delete(flatjoinlist,ver-)
	print(" ", >> llogfile)
# Zero out the Bad Pixels in flatlist
	if (zzeroout) {
		print("===> Zero out the Bad Pixels before making the flat")
	   print("===> Zero out the Bad Pixels before making flat", >> llogfile)
	        imarith("@"//flatimagelist,"*",bbadmask,"@"//flatlist,verbose-)
		hedit("@"//flatlist,"ZEROOUT",bbadmask,del-,add+,ver-,sho-,upd+)
print("IMARITH: @"//flatimagelist//" * "//bbadmask//" @"//flatlist, >> llogfile)
		print(" ", >> llogfile)
	} else {
		imcopy("@"//flatimagelist, "@"//flatlist, verbose-)
		print("IMCOPY: @"//flatimagelist//" @"//flatlist, >> llogfile)
	}
# No dark subtraction
} else {
	if (zzeroout) {
		print("===> Zero out the Bad Pixels before making the flat")
	   print("===> Zero out the Bad Pixels before making flat", >> llogfile)
		imarith("@"//iimagelist, "*", bbadmask, "@"//flatlist, verbose-)
		hedit("@"//flatlist,"ZEROOUT",bbadmask,del-,ad+,ve-,sho-,upd+)
   print("IMARITH: @"//iimagelist//" * "//bbadmask//" @"//flatlist, >> llogfile)
		print(" ", >> llogfile)
	} else {
		print("===> Copying to temporary images")
		imcopy("@"//iimagelist, "@"//flatlist, verbose-)
		print("IMCOPY: @"//iimagelist//" @"//flatlist, >> llogfile)
	}
}

# Normalize each image in flatimagelist/flatmakelist by the bad-pixel-corrected
# (and possibly iterative) mean of its corresponding image in flatlist, and to
# the average bad-pixel-corrected (and possibly iterative) mean of all images 
# in flatlist.  (Images in flatimagelist are original or dark-subtracted;
# images in flatlist are original or dark-subtracted and/or bad-pixel-zeroed.)
meanlist=""
print(" ", >> llogfile); print("===> Normalize each image by its mean (accounting for bad pixels)", >> llogfile)
print("===> Normalize each image by its mean (accounting for bad pixels)")
if(iter_stat){
	if (ddarksub) {
		join (flatimagelist, flatlist, out=flatjoinlist)
	} else {
		join (iimagelist, flatlist, out=flatjoinlist)
	}
	list = flatjoinlist
	while (fscan (list, origimage, currentflat) !=EOF) {
		imgets(currentflat,"itermean",>&"dev$null")
        	if(imgets.value=="0"){
			iterstat(currentflat, nsigrej=5., maxiter=10., print+, ver-, addh+, lower=0.0000001, upper=INDEF, >> llogfile)
		hedit(currentflat,"itermean",iterstat.mean,add+,ver-,sho-,upd+)
		}
	}
	delete(flatjoinlist,ver-)
	hsel("@"//flatlist, "itermean", yes) | average >> "tempffile"
} else {
	imstat("@"//flatlist, lower=0.0000001, >> llogfile)
	imstat("@"//flatlist, fields="mean", lower=0.0000001, format=no) | average >> "tempffile"
}
meaner = average.mean
del ("tempffile", ver-); print(" ", >> llogfile)
if (ddarksub) {
	join (flatimagelist, flatmakelist, out=flatjoinlist)
} else {
	join (iimagelist, flatmakelist, out=flatjoinlist)
}
list = flatjoinlist
while (fscan (list, tempflat, currentflat) !=EOF) {
	meanfile = mktemp("mean")
	meanlist = meanfile
	if(iter_stat){
		hsel(flatbase//currentflat, "itermean", yes, >> meanfile)
	} else {
		imstat(flatbase//currentflat, fields="mean", format=no, lower=0.0000001, >> meanfile)
	}
	j=fscan(meanlist, mean)
	meanest = mean / meaner
	imarith(tempflat, "/", meanest, currentflat, verbose-)
    print("IMARITH: "//currentflat//" / "//meanest//" "//currentflat,>>llogfile)
	meanlist = ""
	delete(meanfile,ver-)
}
delete(flatjoinlist,ver-)
imdel("@"//flatlist, ver-)

# Combine all images to make the flat
print(" ", >> llogfile)
print("===> Combine all images to make the flat")
print("===> Combine all images to make the flat", >> llogfile)
# If there are 100 or more images to combine, stack the images first.
count(flatmakelist, >> "temp.count.thang")
templist="temp.count.thang"
list2=templist
while (fscan (list2, y, z, tempstring, otherstring) !=EOF) {
	if(y>=100) {
		if(v210) {
			imstack("@"//flatmakelist, "stack."//flat)
			combine("stack."//flat, flat, project+, sigma=s_sigma, delete-, clobber-, combine=s_combine, reject=s_reject, outtype="real", offsets=s_offsets, masktype=s_masktyp, maskvalue=s_maskval, blank=s_blank, scale="none", zero=s_zero, weight=s_weight, statsec=s_statsec, lthresh=s_lthresh, hthresh=s_hthresh, nlow=s_nlow, nhigh=s_nhigh, nkeep=s_nkeep, mclip+, lsigma=s_lsigma, hsigma=s_hsigma, rdnoise=s_rdnoise, gain=s_gain, pclip=s_pclip, grow=s_grow)
			imdel("stack."//flat, ver-)
		} else {
			imstack("@"//flatmakelist, "stack."//flat)
			imcombine("stack."//flat, flat, project+, sigma=s_sigma, combine=s_combine, reject=s_reject, offsets=s_offsets, masktype=s_masktyp, maskvalue=s_maskval, blank=s_blank, scale="none", zero=s_zero, weight=s_weight, statsec=s_statsec, expname=s_expname, lthresh=s_lthresh, hthresh=s_hthresh, nlow=s_nlow, nhigh=s_nhigh, nkeep=s_nkeep, mclip+, lsigma=s_lsigma, hsigma=s_hsigma, rdnoise=s_rdnoise, gain=s_gain, pclip=s_pclip, grow=s_grow, logfile='STDOUT', >> llogfile)
			imdel("stack."//flat, ver-)
		}	
	} else {
		if(v210) {
			combine("@"//flatmakelist, flat, sigma=s_sigma, delete-, clobber-, combine=s_combine, reject=s_reject, outtype="real", offsets=s_offsets, masktype=s_masktyp, maskvalue=s_maskval, blank=s_blank, scale="none", zero=s_zero, weight=s_weight, statsec=s_statsec, lthresh=s_lthresh, hthresh=s_hthresh, nlow=s_nlow, nhigh=s_nhigh, nkeep=s_nkeep, mclip+, lsigma=s_lsigma, hsigma=s_hsigma, rdnoise=s_rdnoise, gain=s_gain, pclip=s_pclip, grow=s_grow)
		} else {
			imcombine("@"//flatmakelist, flat, sigma=s_sigma, combine=s_combine, reject=s_reject, offsets=s_offsets, masktype=s_masktyp, maskvalue=s_maskval, blank=s_blank, scale="none", zero=s_zero, weight=s_weight, statsec=s_statsec, expname=s_expname, lthresh=s_lthresh, hthresh=s_hthresh, nlow=s_nlow, nhigh=s_nhigh, nkeep=s_nkeep, mclip+, lsigma=s_lsigma, hsigma=s_hsigma, rdnoise=s_rdnoise, gain=s_gain, pclip=s_pclip, grow=s_grow, logfile='STDOUT',>> llogfile)
		}
	}
}
del("temp.count.thang", ver-)
hedit(flat, "title", "Flatfield created from "//iimagelist, ver-, show-, update+)
if (!ddarksub) imdel("@"//flatmakelist, ver-)

# Divide flat by its mean value, accounting for bad pixels
meanfile = mktemp("mean")
meanlist = meanfile
tempflat = flat//"cameldung"
imarith(flat, "*", bbadmask, tempflat, verbose-)
if(iter_stat){
	iterstat(tempflat, nsigrej=5, maxiter=10, print+, ver-, addh+, lower=0.0000001, upper=INDEF, >> llogfile)
	hsel(tempflat, "itermean", yes, >> meanfile)
} else {
	print(" ",>>llogfile);imstat(tempflat,>>llogfile);print(" ",>>llogfile)
	imstat(tempflat, fields="mean", format=no, lower=0.0000001, >> meanfile)
}
j=fscan(meanlist, mean)
imarith(flat, "/", mean, flat, verbose-)
print("IMARITH: "//flat//" / "//mean//" "//flat, >> llogfile)
print(" ", >>llogfile)
meanlist = ""
delete(meanfile, ver-)
imdelete(tempflat, ver-)

# Zero out Bad Pixels if desired
if (zzeroout) {
	print("===> Zero out bad pixels on flat")
	print("===> Zero out bad pixels on flat", >> llogfile)
	imar(flat,"*",bbadmask,flat,ver-)
	print("IMARITH: "//flat//" * "//bbadmask//" "//flat, >> llogfile)
	hedit(flat,"ZEROOUT",bbadmask,del-,ad+,ve-,sho-,upd+)
	print(" ", >> llogfile)
} else {
	print("===> Bad pixel zeroing not done", >> llogfile)
}

# End of mkflat+ portion
} else {
print("===> "//flat//" already exists -- not recreated -- continuing...")
print(flat//" already exists -- not recreated",>>llogfile);print(" ",>>llogfile)
}

# Divide All Images by the Flat Field
if(ddoflat){
	print("===> Correct All Images for Flat Field")
	print("===> Correct All Images for Flat Field", >> llogfile)
	if (ddarksub) {
imar("@"//flatimagelist,"/",flat,"@"//flatimagelist,ver-,noact-,divzero=0.)
print("IMARITH: @"//flatimagelist//" / "//flat//" @"//flatimagelist,>>llogfile)
	} else {
imar("@"//iimagelist,"/",flat,"@"//flatimagelist,ver-,noact-,divzero=0.)
print("IMARITH: @"//iimagelist//" / "//flat//" @"//flatimagelist,>>llogfile)
	}
	print(" ", >> llogfile)

# Do throughput correction if desired.
if (ttpl) {
	print("===> Divide by Throughput Correction")
	print("===> Divide by Throughput Correction", >> llogfile)
	list = flatimagelist
	while (fscan (list, name) !=EOF) {
		imgets(name,"tplimg",>&"dev$null")
		if(imgets.value=="0"){
			imar(name, "/", ttplname, name, ver-)
			hedit(name,fields="TPL",value="TPL correction made using image"//ttplname,add+,ver-,up+,sho-)
		   hedit(name,fields="TPLIMG",value=ttplname,add+,ver-,up+,sho-)
		  	print(name//" / "//ttplname//" "//name, >> llogfile)
		} else {
		  	print(name//" already throughput-corrected using "//imgets.value, >> llogfile)
	       print(name//" already throughput-corrected using "//imgets.value)
		}
	}
}

# Force updating of pixel values, just to be safe
	print("===> Update pixel values")
	minmax("@"//flatimagelist, force+, update+, verbose-)
} else {
	print("===> Flat fielding of images not done", >> llogfile)
}

print("===> Mopping Up...")
del(flatlist, ver-, >& "dev$null"); del(flatmakelist, ver-, >& "dev$null")
del(flatbase, ver-, >& "dev$null")
datefile = mktemp("date")
date(>>datefile)
datelist = datefile
print("finished at "//datelist, >> llogfile)
datelist = ""
del(datefile, ver-)
print("...it is accomplished.")
} else {
	beep
        print("WARNING: package noao.imred.ccdred is not loaded!")
        print("Please load that package and restart this task.")
}

if(echog)beep

end

